use nanoneo::{
    core::element::{Element, Render},
    text, read_param,
};
use pathbufd::PathBufD;
use regex::Regex;
use std::{
    collections::HashMap,
    fs::{exists, read_to_string, write},
    sync::LazyLock,
    time::SystemTime,
};
use tera::Context;
use tetratto_core::{
    config::Config,
    model::{
        auth::{DefaultTimelineChoice, User},
        permissions::{FinePermission, SecondaryPermission},
    },
};
use tetratto_l10n::LangFile;
use tetratto_shared::hash::salt;
use tokio::sync::RwLock;
use crate::{create_dir_if_not_exists, write_if_track, write_template};

// images
pub const DEFAULT_AVATAR: &str = include_str!("./public/images/default-avatar.svg");
pub const DEFAULT_BANNER: &str = include_str!("./public/images/default-banner.svg");
pub const FAVICON: &str = include_str!("./public/images/favicon.svg");

// css
pub const STYLE_CSS: &str = include_str!("./public/css/style.css");
pub const ROOT_CSS: &str = include_str!("./public/css/root.css");
pub const UTILITY_CSS: &str = include_str!("./public/css/utility.css");
pub const CHATS_CSS: &str = include_str!("./public/css/chats.css");

// js
pub const LOADER_JS: &str = include_str!("./public/js/loader.js");
pub const ATTO_JS: &str = include_str!("./public/js/atto.js");
pub const ME_JS: &str = include_str!("./public/js/me.js");
pub const STREAMS_JS: &str = include_str!("./public/js/streams.js");
pub const CARP_JS: &str = include_str!("./public/js/carp.js");
pub const PROTO_LINKS_JS: &str = include_str!("./public/js/proto_links.js");
pub const APP_SDK_JS: &str = include_str!("./public/js/app_sdk.js");

// html
pub const BODY: &str = include_str!("./public/html/body.lisp");
pub const ROOT: &str = include_str!("./public/html/root.lisp");
pub const MACROS: &str = include_str!("./public/html/macros.lisp");
pub const COMPONENTS: &str = include_str!("./public/html/components.lisp");

pub const MISC_ERROR: &str = include_str!("./public/html/misc/error.lisp");
pub const MISC_NOTIFICATIONS: &str = include_str!("./public/html/misc/notifications.lisp");
pub const MISC_MARKDOWN: &str = include_str!("./public/html/misc/markdown.lisp");
pub const MISC_REQUESTS: &str = include_str!("./public/html/misc/requests.lisp");
pub const MISC_ACHIEVEMENTS: &str = include_str!("./public/html/misc/achievements.lisp");

pub const AUTH_BASE: &str = include_str!("./public/html/auth/base.lisp");
pub const AUTH_LOGIN: &str = include_str!("./public/html/auth/login.lisp");
pub const AUTH_REGISTER: &str = include_str!("./public/html/auth/register.lisp");
pub const AUTH_CONNECTION: &str = include_str!("./public/html/auth/connection.lisp");
pub const AUTH_SELLER_CONNECTION: &str = include_str!("./public/html/auth/seller_connection.lisp");

pub const PROFILE_BASE: &str = include_str!("./public/html/profile/base.lisp");
pub const PROFILE_POSTS: &str = include_str!("./public/html/profile/posts.lisp");
pub const PROFILE_SETTINGS: &str = include_str!("./public/html/profile/settings.lisp");
pub const PROFILE_FOLLOWING: &str = include_str!("./public/html/profile/following.lisp");
pub const PROFILE_FOLLOWERS: &str = include_str!("./public/html/profile/followers.lisp");
pub const PROFILE_WARNING: &str = include_str!("./public/html/profile/warning.lisp");
pub const PROFILE_PRIVATE: &str = include_str!("./public/html/profile/private.lisp");
pub const PROFILE_BLOCKED: &str = include_str!("./public/html/profile/blocked.lisp");
pub const PROFILE_BANNED: &str = include_str!("./public/html/profile/banned.lisp");
pub const PROFILE_REPLIES: &str = include_str!("./public/html/profile/replies.lisp");
pub const PROFILE_MEDIA: &str = include_str!("./public/html/profile/media.lisp");
pub const PROFILE_OUTBOX: &str = include_str!("./public/html/profile/outbox.lisp");
pub const PROFILE_RESPONSES: &str = include_str!("./public/html/profile/responses.lisp");

pub const COMMUNITIES_LIST: &str = include_str!("./public/html/communities/list.lisp");
pub const COMMUNITIES_BASE: &str = include_str!("./public/html/communities/base.lisp");
pub const COMMUNITIES_FEED: &str = include_str!("./public/html/communities/feed.lisp");
pub const COMMUNITIES_SETTINGS: &str = include_str!("./public/html/communities/settings.lisp");
pub const COMMUNITIES_MEMBERS: &str = include_str!("./public/html/communities/members.lisp");
pub const COMMUNITIES_SEARCH: &str = include_str!("./public/html/communities/search.lisp");
pub const COMMUNITIES_CREATE_POST: &str =
    include_str!("./public/html/communities/create_post.lisp");
pub const COMMUNITIES_QUESTION: &str = include_str!("./public/html/communities/question.lisp");
pub const COMMUNITIES_QUESTIONS: &str = include_str!("./public/html/communities/questions.lisp");

pub const POST_POST: &str = include_str!("./public/html/post/post.lisp");
pub const POST_REPOSTS: &str = include_str!("./public/html/post/reposts.lisp");
pub const POST_QUOTES: &str = include_str!("./public/html/post/quotes.lisp");
pub const POST_LIKES: &str = include_str!("./public/html/post/likes.lisp");

pub const TIMELINES_HOME: &str = include_str!("./public/html/timelines/home.lisp");
pub const TIMELINES_POPULAR: &str = include_str!("./public/html/timelines/popular.lisp");
pub const TIMELINES_FOLLOWING: &str = include_str!("./public/html/timelines/following.lisp");
pub const TIMELINES_ALL: &str = include_str!("./public/html/timelines/all.lisp");
pub const TIMELINES_HOME_QUESTIONS: &str =
    include_str!("./public/html/timelines/home_questions.lisp");
pub const TIMELINES_POPULAR_QUESTIONS: &str =
    include_str!("./public/html/timelines/popular_questions.lisp");
pub const TIMELINES_FOLLOWING_QUESTIONS: &str =
    include_str!("./public/html/timelines/following_questions.lisp");
pub const TIMELINES_ALL_QUESTIONS: &str =
    include_str!("./public/html/timelines/all_questions.lisp");
pub const TIMELINES_SEARCH: &str = include_str!("./public/html/timelines/search.lisp");
pub const TIMELINES_SWISS_ARMY: &str = include_str!("./public/html/timelines/swiss_army.lisp");

pub const MOD_AUDIT_LOG: &str = include_str!("./public/html/mod/audit_log.lisp");
pub const MOD_REPORTS: &str = include_str!("./public/html/mod/reports.lisp");
pub const MOD_FILE_REPORT: &str = include_str!("./public/html/mod/file_report.lisp");
pub const MOD_IP_BANS: &str = include_str!("./public/html/mod/ip_bans.lisp");
pub const MOD_PROFILE: &str = include_str!("./public/html/mod/profile.lisp");
pub const MOD_WARNINGS: &str = include_str!("./public/html/mod/warnings.lisp");
pub const MOD_STATS: &str = include_str!("./public/html/mod/stats.lisp");

pub const CHATS_APP: &str = include_str!("./public/html/chats/app.lisp");
pub const CHATS_STREAM: &str = include_str!("./public/html/chats/stream.lisp");
pub const CHATS_MESSAGE: &str = include_str!("./public/html/chats/message.lisp");
pub const CHATS_CHANNELS: &str = include_str!("./public/html/chats/channels.lisp");

pub const STACKS_LIST: &str = include_str!("./public/html/stacks/list.lisp");
pub const STACKS_FEED: &str = include_str!("./public/html/stacks/feed.lisp");
pub const STACKS_MANAGE: &str = include_str!("./public/html/stacks/manage.lisp");
pub const STACKS_ADD_USER: &str = include_str!("./public/html/stacks/add_user.lisp");

pub const FORGE_HOME: &str = include_str!("./public/html/forge/home.lisp");
pub const FORGE_BASE: &str = include_str!("./public/html/forge/base.lisp");
pub const FORGE_INFO: &str = include_str!("./public/html/forge/info.lisp");
pub const FORGE_TICKETS: &str = include_str!("./public/html/forge/tickets.lisp");

pub const DEVELOPER_HOME: &str = include_str!("./public/html/developer/home.lisp");
pub const DEVELOPER_APP: &str = include_str!("./public/html/developer/app.lisp");
pub const DEVELOPER_LINK: &str = include_str!("./public/html/developer/link.lisp");

pub const JOURNALS_APP: &str = include_str!("./public/html/journals/app.lisp");

pub const LITTLEWEB_SERVICES: &str = include_str!("./public/html/littleweb/services.lisp");
pub const LITTLEWEB_DOMAINS: &str = include_str!("./public/html/littleweb/domains.lisp");
pub const LITTLEWEB_SERVICE: &str = include_str!("./public/html/littleweb/service.lisp");
pub const LITTLEWEB_DOMAIN: &str = include_str!("./public/html/littleweb/domain.lisp");
pub const LITTLEWEB_BROWSER: &str = include_str!("./public/html/littleweb/browser.lisp");

pub const MARKETPLACE_SELLER: &str = include_str!("./public/html/marketplace/seller.lisp");

// langs
pub const LANG_EN_US: &str = include_str!("./langs/en-US.toml");

// ...

pub const VENDOR_SPOTIFY_ICON: &str = include_str!("./public/images/vendor/spotify.svg");
pub const VENDOR_LAST_FM_ICON: &str = include_str!("./public/images/vendor/last-fm.svg");
pub const VENDOR_STRIPE_ICON: &str = include_str!("./public/images/vendor/stripe.svg");

pub const TETRATTO_BUNNY: &[u8] = include_bytes!("./public/images/tetratto_bunny.webp");

pub(crate) static HTML_FOOTER: LazyLock<RwLock<String>> =
    LazyLock::new(|| RwLock::new(String::new()));

/// A container for all loaded icons.
pub(crate) static ICONS: LazyLock<RwLock<HashMap<String, String>>> =
    LazyLock::new(|| RwLock::new(HashMap::new()));

/// Pull an icon given its name and insert it into [`ICONS`].
pub(crate) async fn pull_icon(icon: &str, icons_dir: &str) {
    let writer = &mut ICONS.write().await;

    let icon_url = format!(
        "https://raw.githubusercontent.com/lucide-icons/lucide/refs/heads/main/icons/{icon}.svg"
    );

    let file_path = PathBufD::current().extend(&[icons_dir, &format!("{icon}.svg")]);

    if exists(&file_path).unwrap() {
        writer.insert(icon.to_string(), read_to_string(&file_path).unwrap());
        return;
    }

    println!("download icon: {icon}");
    let svg = reqwest::get(icon_url)
        .await
        .unwrap()
        .text()
        .await
        .unwrap()
        .replace("\n", "");

    write(&file_path, &svg).unwrap();
    writer.insert(icon.to_string(), svg);
}

macro_rules! vendor_icon {
    ($name:literal, $icon:ident, $icons_dir:expr) => {{
        let writer = &mut ICONS.write().await;
        writer.insert($name.to_string(), $icon.to_string());

        let file_path =
            PathBufD::current().extend(&[$icons_dir.clone(), format!("{}.svg", $name.to_string())]);
        std::fs::write(file_path, $icon).unwrap();
    }};
}

macro_rules! bin_icon {
    ($name:literal, $icon:ident, $icons_dir:expr) => {{
        let file_path = PathBufD::current().extend(&[$icons_dir.clone(), $name.to_string()]);
        std::fs::write(file_path, $icon).unwrap();
    }};
}

/// Read a string and replace all custom blocks with the corresponding correct HTML.
///
/// # Replaces
/// * icons
/// * icons (with class specifier)
/// * l10n text
pub(crate) async fn replace_in_html(
    input: &str,
    config: &Config,
    lisp: bool,
    plugins: Option<&mut HashMap<String, Box<dyn FnMut(Element) -> Element>>>,
) -> String {
    let reader = HTML_FOOTER.read().await;

    if reader.is_empty() {
        if !config.html_footer_path.is_empty() {
            drop(reader);

            let mut writer = HTML_FOOTER.write().await;
            *writer = read_to_string(PathBufD::current().join(&config.html_footer_path)).unwrap();
        } else {
            drop(reader);
        }
    } else {
        drop(reader);
    }

    let reader = HTML_FOOTER.read().await;

    // ...
    let mut input = if !lisp {
        input.to_string()
    } else {
        let start = SystemTime::now();
        let parsed = nanoneo::parse(input);
        println!("parsed lisp in {}μs", start.elapsed().unwrap().as_micros());

        if let Some(plugins) = plugins {
            parsed.render(plugins)
        } else {
            parsed.render_safe()
        }
    };

    input = input.replace("<!-- prettier-ignore -->", "");

    // l10n text
    let text = Regex::new("(\\{\\{)\\s*(text)\\s*\"(.*?)\"\\s*(\\}\\})").unwrap();

    for cap in text.captures_iter(&input.clone()) {
        let replace_with = format!("{{{{ lang[\"{}\"] }}}}", cap.get(3).unwrap().as_str());
        input = input.replace(cap.get(0).unwrap().as_str(), &replace_with);
    }

    // icon (with class)
    let icon_with_class =
        Regex::new("(\\{\\{)\\s*(icon)\\s*(.*?)\\s*c\\((.*?)\\)\\s*(\\}\\})").unwrap();

    for cap in icon_with_class.captures_iter(&input.clone()) {
        let cap_str = &cap.get(3).unwrap().as_str().replace("\"", "");
        let icon = &(if cap_str.contains(" }}") {
            cap_str.split(" }}").next().unwrap().to_string()
        } else {
            cap_str.to_string()
        });

        pull_icon(icon, &config.dirs.icons).await;

        let reader = ICONS.read().await;
        let icon_text = reader.get(icon).unwrap().replace(
            "<svg",
            &format!("<svg class=\"icon {}\"", cap.get(4).unwrap().as_str()),
        );

        input = input.replace(
            &format!(
                "{{{{ icon \"{cap_str}\" c({}) }}}}",
                cap.get(4).unwrap().as_str()
            ),
            &icon_text,
        );
    }

    // icon (without class)
    let icon_without_class = Regex::new("(\\{\\{)\\s*(icon)\\s*(.*?)\\s*(\\}\\})").unwrap();

    for cap in icon_without_class.captures_iter(&input.clone()) {
        let cap_str = &cap.get(3).unwrap().as_str().replace("\"", "");
        let icon = &(if cap_str.contains(" }}") {
            cap_str.split(" }}").next().unwrap().to_string()
        } else {
            cap_str.to_string()
        });

        pull_icon(icon, &config.dirs.icons).await;

        let reader = ICONS.read().await;
        let icon_text = reader
            .get(icon)
            .unwrap()
            .replace("<svg", "<svg class=\"icon\"");

        input = input.replace(&format!("{{{{ icon \"{cap_str}\" }}}}",), &icon_text);
    }

    // return
    input = input.replacen("<!-- html_footer_goes_here -->", &format!("{reader}"), 1);
    input
}

pub(crate) fn lisp_plugins() -> HashMap<String, Box<dyn FnMut(Element) -> Element>> {
    let mut plugins = HashMap::new();

    plugins.insert(
        "icon".to_string(),
        Box::new(|e: Element| text!(format!("\n{{{{ icon \"{}\" }}}}\n", read_param!(e, 0)))) as _,
    );

    plugins.insert(
        "icon_class".to_string(),
        Box::new(|e: Element| {
            text!(format!(
                "\n{{{{ icon \"{}\" c({}) }}}}\n",
                read_param!(e, 0),
                read_param!(e, 1)
            ))
        }) as _,
    );

    plugins.insert(
        "str".to_string(),
        Box::new(|e: Element| text!(format!("{{{{ text \"{}\" }}}}", read_param!(e, 0)))) as _,
    );

    plugins
}

/// Set up public directories.
pub(crate) async fn write_assets(config: &Config) -> PathBufD {
    vendor_icon!("spotify", VENDOR_SPOTIFY_ICON, config.dirs.icons);
    vendor_icon!("last_fm", VENDOR_LAST_FM_ICON, config.dirs.icons);
    vendor_icon!("stripe", VENDOR_STRIPE_ICON, config.dirs.icons);
    bin_icon!("tetratto_bunny.webp", TETRATTO_BUNNY, config.dirs.assets);

    // ...
    let mut plugins = lisp_plugins();
    let html_path = PathBufD::current().join(&config.dirs.templates);

    write_template!(html_path->"body.html"(crate::assets::BODY) --config=config --lisp plugins);
    write_template!(html_path->"root.html"(crate::assets::ROOT) --config=config --lisp plugins);
    write_template!(html_path->"macros.html"(crate::assets::MACROS) --config=config --lisp plugins);
    write_template!(html_path->"components.html"(crate::assets::COMPONENTS) --config=config --lisp plugins);

    write_template!(html_path->"misc/error.html"(crate::assets::MISC_ERROR) -d "misc" --config=config --lisp plugins);
    write_template!(html_path->"misc/notifications.html"(crate::assets::MISC_NOTIFICATIONS) --config=config --lisp plugins);
    write_template!(html_path->"misc/markdown.html"(crate::assets::MISC_MARKDOWN) --config=config --lisp plugins);
    write_template!(html_path->"misc/requests.html"(crate::assets::MISC_REQUESTS) --config=config --lisp plugins);
    write_template!(html_path->"misc/achievements.html"(crate::assets::MISC_ACHIEVEMENTS) --config=config --lisp plugins);

    write_template!(html_path->"auth/base.html"(crate::assets::AUTH_BASE) -d "auth" --config=config --lisp plugins);
    write_template!(html_path->"auth/login.html"(crate::assets::AUTH_LOGIN) --config=config --lisp plugins);
    write_template!(html_path->"auth/register.html"(crate::assets::AUTH_REGISTER) --config=config --lisp plugins);
    write_template!(html_path->"auth/connection.html"(crate::assets::AUTH_CONNECTION) --config=config --lisp plugins);
    write_template!(html_path->"auth/seller_connection.html"(crate::assets::AUTH_SELLER_CONNECTION) --config=config --lisp plugins);

    write_template!(html_path->"profile/base.html"(crate::assets::PROFILE_BASE) -d "profile" --config=config --lisp plugins);
    write_template!(html_path->"profile/posts.html"(crate::assets::PROFILE_POSTS) --config=config --lisp plugins);
    write_template!(html_path->"profile/settings.html"(crate::assets::PROFILE_SETTINGS) --config=config --lisp plugins);
    write_template!(html_path->"profile/following.html"(crate::assets::PROFILE_FOLLOWING) --config=config --lisp plugins);
    write_template!(html_path->"profile/followers.html"(crate::assets::PROFILE_FOLLOWERS) --config=config --lisp plugins);
    write_template!(html_path->"profile/warning.html"(crate::assets::PROFILE_WARNING) --config=config --lisp plugins);
    write_template!(html_path->"profile/private.html"(crate::assets::PROFILE_PRIVATE) --config=config --lisp plugins);
    write_template!(html_path->"profile/blocked.html"(crate::assets::PROFILE_BLOCKED) --config=config --lisp plugins);
    write_template!(html_path->"profile/banned.html"(crate::assets::PROFILE_BANNED) --config=config --lisp plugins);
    write_template!(html_path->"profile/replies.html"(crate::assets::PROFILE_REPLIES) --config=config --lisp plugins);
    write_template!(html_path->"profile/media.html"(crate::assets::PROFILE_MEDIA) --config=config --lisp plugins);
    write_template!(html_path->"profile/outbox.html"(crate::assets::PROFILE_OUTBOX) --config=config --lisp plugins);
    write_template!(html_path->"profile/responses.html"(crate::assets::PROFILE_RESPONSES) --config=config --lisp plugins);

    write_template!(html_path->"communities/list.html"(crate::assets::COMMUNITIES_LIST) -d "communities" --config=config --lisp plugins);
    write_template!(html_path->"communities/base.html"(crate::assets::COMMUNITIES_BASE) --config=config --lisp plugins);
    write_template!(html_path->"communities/feed.html"(crate::assets::COMMUNITIES_FEED) --config=config --lisp plugins);
    write_template!(html_path->"communities/settings.html"(crate::assets::COMMUNITIES_SETTINGS) --config=config --lisp plugins);
    write_template!(html_path->"communities/members.html"(crate::assets::COMMUNITIES_MEMBERS) --config=config --lisp plugins);
    write_template!(html_path->"communities/search.html"(crate::assets::COMMUNITIES_SEARCH) --config=config --lisp plugins);
    write_template!(html_path->"communities/create_post.html"(crate::assets::COMMUNITIES_CREATE_POST) --config=config --lisp plugins);
    write_template!(html_path->"communities/question.html"(crate::assets::COMMUNITIES_QUESTION) --config=config --lisp plugins);
    write_template!(html_path->"communities/questions.html"(crate::assets::COMMUNITIES_QUESTIONS) --config=config --lisp plugins);

    write_template!(html_path->"post/post.html"(crate::assets::POST_POST) -d "post" --config=config --lisp plugins);
    write_template!(html_path->"post/reposts.html"(crate::assets::POST_REPOSTS) --config=config --lisp plugins);
    write_template!(html_path->"post/quotes.html"(crate::assets::POST_QUOTES) --config=config --lisp plugins);
    write_template!(html_path->"post/likes.html"(crate::assets::POST_LIKES) --config=config --lisp plugins);

    write_template!(html_path->"timelines/home.html"(crate::assets::TIMELINES_HOME) -d "timelines" --config=config --lisp plugins);
    write_template!(html_path->"timelines/popular.html"(crate::assets::TIMELINES_POPULAR) --config=config --lisp plugins);
    write_template!(html_path->"timelines/following.html"(crate::assets::TIMELINES_FOLLOWING) --config=config --lisp plugins);
    write_template!(html_path->"timelines/all.html"(crate::assets::TIMELINES_ALL) --config=config --lisp plugins);
    write_template!(html_path->"timelines/home_questions.html"(crate::assets::TIMELINES_HOME_QUESTIONS) --config=config --lisp plugins);
    write_template!(html_path->"timelines/popular_questions.html"(crate::assets::TIMELINES_POPULAR_QUESTIONS) --config=config --lisp plugins);
    write_template!(html_path->"timelines/following_questions.html"(crate::assets::TIMELINES_FOLLOWING_QUESTIONS) --config=config --lisp plugins);
    write_template!(html_path->"timelines/all_questions.html"(crate::assets::TIMELINES_ALL_QUESTIONS) --config=config --lisp plugins);
    write_template!(html_path->"timelines/search.html"(crate::assets::TIMELINES_SEARCH) --config=config --lisp plugins);
    write_template!(html_path->"timelines/swiss_army.html"(crate::assets::TIMELINES_SWISS_ARMY) --config=config --lisp plugins);

    write_template!(html_path->"mod/audit_log.html"(crate::assets::MOD_AUDIT_LOG) -d "mod" --config=config --lisp plugins);
    write_template!(html_path->"mod/reports.html"(crate::assets::MOD_REPORTS) --config=config --lisp plugins);
    write_template!(html_path->"mod/file_report.html"(crate::assets::MOD_FILE_REPORT) --config=config --lisp plugins);
    write_template!(html_path->"mod/ip_bans.html"(crate::assets::MOD_IP_BANS) --config=config --lisp plugins);
    write_template!(html_path->"mod/profile.html"(crate::assets::MOD_PROFILE) --config=config --lisp plugins);
    write_template!(html_path->"mod/warnings.html"(crate::assets::MOD_WARNINGS) --config=config --lisp plugins);
    write_template!(html_path->"mod/stats.html"(crate::assets::MOD_STATS) --config=config --lisp plugins);

    write_template!(html_path->"chats/app.html"(crate::assets::CHATS_APP) -d "chats" --config=config --lisp plugins);
    write_template!(html_path->"chats/stream.html"(crate::assets::CHATS_STREAM) --config=config --lisp plugins);
    write_template!(html_path->"chats/message.html"(crate::assets::CHATS_MESSAGE) --config=config --lisp plugins);
    write_template!(html_path->"chats/channels.html"(crate::assets::CHATS_CHANNELS) --config=config --lisp plugins);

    write_template!(html_path->"stacks/list.html"(crate::assets::STACKS_LIST) -d "stacks" --config=config --lisp plugins);
    write_template!(html_path->"stacks/feed.html"(crate::assets::STACKS_FEED) --config=config --lisp plugins);
    write_template!(html_path->"stacks/manage.html"(crate::assets::STACKS_MANAGE) --config=config --lisp plugins);
    write_template!(html_path->"stacks/add_user.html"(crate::assets::STACKS_ADD_USER) --config=config --lisp plugins);

    write_template!(html_path->"forge/home.html"(crate::assets::FORGE_HOME) -d "forge" --config=config --lisp plugins);
    write_template!(html_path->"forge/base.html"(crate::assets::FORGE_BASE) --config=config --lisp plugins);
    write_template!(html_path->"forge/info.html"(crate::assets::FORGE_INFO) --config=config --lisp plugins);
    write_template!(html_path->"forge/tickets.html"(crate::assets::FORGE_TICKETS) --config=config --lisp plugins);

    write_template!(html_path->"developer/home.html"(crate::assets::DEVELOPER_HOME) -d "developer" --config=config --lisp plugins);
    write_template!(html_path->"developer/app.html"(crate::assets::DEVELOPER_APP) --config=config --lisp plugins);
    write_template!(html_path->"developer/link.html"(crate::assets::DEVELOPER_LINK) --config=config --lisp plugins);

    write_template!(html_path->"journals/app.html"(crate::assets::JOURNALS_APP) -d "journals" --config=config --lisp plugins);

    write_template!(html_path->"littleweb/services.html"(crate::assets::LITTLEWEB_SERVICES) -d "littleweb" --config=config --lisp plugins);
    write_template!(html_path->"littleweb/domains.html"(crate::assets::LITTLEWEB_DOMAINS) --config=config --lisp plugins);
    write_template!(html_path->"littleweb/service.html"(crate::assets::LITTLEWEB_SERVICE) --config=config --lisp plugins);
    write_template!(html_path->"littleweb/domain.html"(crate::assets::LITTLEWEB_DOMAIN) --config=config --lisp plugins);
    write_template!(html_path->"littleweb/browser.html"(crate::assets::LITTLEWEB_BROWSER) --config=config --lisp plugins);

    write_template!(html_path->"marketplace/seller.html"(crate::assets::MARKETPLACE_SELLER) -d "marketplace" --config=config --lisp plugins);

    html_path
}

/// Set up extra directories.
pub(crate) async fn init_dirs(config: &Config) {
    create_dir_if_not_exists!(&config.dirs.templates);
    create_dir_if_not_exists!(&config.dirs.docs);
    create_dir_if_not_exists!(&config.dirs.rustdoc);
    create_dir_if_not_exists!(&config.dirs.assets);

    // images
    create_dir_if_not_exists!(&config.dirs.media);
    let images_path = PathBufD::current().extend(&[config.dirs.media.as_str(), "images"]);
    create_dir_if_not_exists!(&images_path);
    create_dir_if_not_exists!(
        &PathBufD::current().extend(&[config.dirs.media.as_str(), "avatars"])
    );
    create_dir_if_not_exists!(
        &PathBufD::current().extend(&[config.dirs.media.as_str(), "community_avatars"])
    );
    create_dir_if_not_exists!(
        &PathBufD::current().extend(&[config.dirs.media.as_str(), "banners"])
    );
    create_dir_if_not_exists!(
        &PathBufD::current().extend(&[config.dirs.media.as_str(), "community_banners"])
    );
    create_dir_if_not_exists!(
        &PathBufD::current().extend(&[config.dirs.media.as_str(), "uploads"])
    );

    write_if_track!(images_path->"default-avatar.svg"(DEFAULT_AVATAR) --config=config);
    write_if_track!(images_path->"default-banner.svg"(DEFAULT_BANNER) --config=config);
    write_if_track!(images_path->"favicon.svg"(FAVICON) --config=config);

    // icons
    create_dir_if_not_exists!(&PathBufD::current().join(config.dirs.icons.as_str()));

    // langs
    let langs_path = PathBufD::current().join("langs");
    create_dir_if_not_exists!(&langs_path);

    write_template!(langs_path->"en-US.toml"(LANG_EN_US));
}

/// A random ASCII value inserted into the URL of static assets to "break" the cache. Essentially just for cache busting.
pub(crate) static CACHE_BREAKER: LazyLock<String> = LazyLock::new(salt);

/// Create the initial template context.
pub(crate) async fn initial_context(
    config: &Config,
    lang: &LangFile,
    user: &Option<User>,
) -> Context {
    let mut ctx = Context::new();
    ctx.insert("config", &config);
    ctx.insert("pubsub", &true);

    ctx.insert("user", &user);
    ctx.insert("use_user_theme", &true);

    if let Some(ua) = user {
        ctx.insert("is_helper", &ua.permissions.check_helper());
        ctx.insert("is_manager", &ua.permissions.check_manager());
        ctx.insert(
            "is_supporter",
            &ua.permissions.check(FinePermission::SUPPORTER),
        );
        ctx.insert(
            "has_developer_pass",
            &ua.secondary_permissions
                .check(SecondaryPermission::DEVELOPER_PASS),
        );
        ctx.insert("home", &ua.settings.default_timeline.relative_url());
    } else {
        ctx.insert("is_helper", &false);
        ctx.insert("is_manager", &false);
        ctx.insert("home", &DefaultTimelineChoice::default().relative_url());
    }

    ctx.insert("lang", &lang.data);
    ctx.insert(
        "random_cache_breaker",
        &if let Ok(c) = std::env::var("CACHE_BREAKER") {
            c
        } else {
            CACHE_BREAKER.clone()
        },
    );
    ctx
}
