use crate::{
    get_app_from_key,
    routes::api::v1::{InsertAppData, QueryAppData, UpdateAppDataValue, UpdateAppDataKey},
    State,
};
use axum::{extract::Path, http::HeaderMap, response::IntoResponse, Extension, Json};
use tetratto_core::model::{
    apps::{AppData, AppDataQuery, AppDataQueryResult},
    ApiReturn, Error,
};

pub async fn get_app_request(
    headers: HeaderMap,
    Extension(data): Extension<State>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let app = match get_app_from_key!(data, headers) {
        Some(x) => x,
        None => return Json(Error::NotAllowed.into()),
    };

    Json(ApiReturn {
        ok: true,
        message: "Success".to_string(),
        payload: Some(app),
    })
}

pub async fn query_request(
    headers: HeaderMap,
    Extension(data): Extension<State>,
    Json(req): Json<QueryAppData>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let app = match get_app_from_key!(data, headers) {
        Some(x) => x,
        None => return Json(Error::NotAllowed.into()),
    };

    match data
        .query_app_data(AppDataQuery {
            app: app.id,
            query: req.query,
            mode: req.mode,
        })
        .await
    {
        Ok(x) => Json(ApiReturn {
            ok: true,
            message: "Success".to_string(),
            payload: Some(x),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn create_request(
    headers: HeaderMap,
    Extension(data): Extension<State>,
    Json(req): Json<InsertAppData>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let app = match get_app_from_key!(data, headers) {
        Some(x) => x,
        None => return Json(Error::NotAllowed.into()),
    };

    let owner = match data.get_user_by_id(app.owner).await {
        Ok(x) => x,
        Err(e) => return Json(e.into()),
    };

    // check size
    let new_size = app.data_used + req.value.len();
    if new_size > AppData::user_limit(&owner) {
        return Json(Error::AppHitStorageLimit.into());
    }

    // ...
    if let Err(e) = data.add_app_data_used(app.id, req.value.len() as i32).await {
        return Json(e.into());
    }

    match data
        .create_app_data(AppData::new(app.id, req.key, req.value))
        .await
    {
        Ok(s) => Json(ApiReturn {
            ok: true,
            message: "Data inserted".to_string(),
            payload: s.id.to_string(),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn update_key_request(
    headers: HeaderMap,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
    Json(req): Json<UpdateAppDataKey>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let app = match get_app_from_key!(data, headers) {
        Some(x) => x,
        None => return Json(Error::NotAllowed.into()),
    };

    let app_data = match data.get_app_data_by_id(id).await {
        Ok(x) => x,
        Err(e) => return Json(e.into()),
    };

    if app_data.app != app.id {
        return Json(Error::NotAllowed.into());
    }

    match data.update_app_data_key(id, &req.key).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Data updated".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn update_value_request(
    headers: HeaderMap,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
    Json(req): Json<UpdateAppDataValue>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let app = match get_app_from_key!(data, headers) {
        Some(x) => x,
        None => return Json(Error::NotAllowed.into()),
    };

    let owner = match data.get_user_by_id(app.owner).await {
        Ok(x) => x,
        Err(e) => return Json(e.into()),
    };

    let app_data = match data.get_app_data_by_id(id).await {
        Ok(x) => x,
        Err(e) => return Json(e.into()),
    };

    if app_data.app != app.id {
        return Json(Error::NotAllowed.into());
    }

    // check size
    let size_without = app.data_used - app_data.value.len();
    let new_size = size_without + req.value.len();

    if new_size > AppData::user_limit(&owner) {
        return Json(Error::AppHitStorageLimit.into());
    }

    // ...
    // we only need to add the delta size (the next size - the old size)
    if let Err(e) = data
        .add_app_data_used(
            app.id,
            (req.value.len() as i32) - (app_data.value.len() as i32),
        )
        .await
    {
        return Json(e.into());
    }

    match data.update_app_data_value(id, &req.value).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Data updated".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn delete_request(
    headers: HeaderMap,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let app = match get_app_from_key!(data, headers) {
        Some(x) => x,
        None => return Json(Error::NotAllowed.into()),
    };

    let app_data = match data.get_app_data_by_id(id).await {
        Ok(x) => x,
        Err(e) => return Json(e.into()),
    };

    if app_data.app != app.id {
        return Json(Error::NotAllowed.into());
    }

    // ...
    if let Err(e) = data
        .add_app_data_used(app.id, -(app_data.value.len() as i32))
        .await
    {
        return Json(e.into());
    }

    match data.delete_app_data(id).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Data deleted".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn delete_query_request(
    headers: HeaderMap,
    Extension(data): Extension<State>,
    Json(req): Json<QueryAppData>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let app = match get_app_from_key!(data, headers) {
        Some(x) => x,
        None => return Json(Error::NotAllowed.into()),
    };

    // ...
    let rows = match data
        .query_app_data(AppDataQuery {
            app: app.id,
            query: req.query.clone(),
            mode: req.mode.clone(),
        })
        .await
    {
        Ok(x) => match x {
            AppDataQueryResult::One(x) => vec![x],
            AppDataQueryResult::Many(x) => x,
        },
        Err(e) => return Json(e.into()),
    };

    let mut subtract_amount: usize = 0;
    for row in &rows {
        subtract_amount += row.value.len();
    }
    drop(rows);

    if let Err(e) = data
        .add_app_data_used(app.id, -(subtract_amount as i32))
        .await
    {
        return Json(e.into());
    }

    match data
        .query_delete_app_data(AppDataQuery {
            app: app.id,
            query: req.query,
            mode: req.mode,
        })
        .await
    {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Data deleted".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}
