use crate::{get_user_from_token, routes::api::v1::CreateMessageReaction, State};
use axum::{Extension, Json, extract::Path, response::IntoResponse};
use crate::cookie::CookieJar;
use tetratto_core::model::{channels::MessageReaction, oauth, ApiReturn, Error};

pub async fn get_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserReact) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data
        .get_message_reactions_by_owner_message(user.id, id)
        .await
    {
        Ok(r) => Json(ApiReturn {
            ok: true,
            message: "Reactions exists".to_string(),
            payload: Some(r),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn create_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Json(req): Json<CreateMessageReaction>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserReact) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    let message_id = match req.message.parse::<usize>() {
        Ok(n) => n,
        Err(e) => return Json(Error::MiscError(e.to_string()).into()),
    };

    // check for existing reaction
    if let Ok(r) = data
        .get_message_reaction_by_owner_message_emoji(user.id, message_id, &req.emoji)
        .await
    {
        if let Err(e) = data.delete_message_reaction(r.id, &user).await {
            return Json(e.into());
        } else {
            return Json(ApiReturn {
                ok: true,
                message: "Reaction removed".to_string(),
                payload: (),
            });
        }
    }

    // create reaction
    match data
        .create_message_reaction(MessageReaction::new(user.id, message_id, req.emoji), &user)
        .await
    {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Reaction created".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn delete_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path((id, emoji)): Path<(usize, String)>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserReact) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    let reaction = match data
        .get_message_reaction_by_owner_message_emoji(user.id, id, &emoji)
        .await
    {
        Ok(r) => r,
        Err(e) => return Json(e.into()),
    };

    match data.delete_message_reaction(reaction.id, &user).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Reaction deleted".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}
