use axum::{
    extract::{Path, Query},
    response::IntoResponse,
    Extension, Json,
};
use crate::cookie::CookieJar;
use tetratto_core::model::{auth::AchievementName, communities::PostDraft, oauth, ApiReturn, Error};
use crate::{
    get_user_from_token,
    routes::{
        api::v1::{CreatePostDraft, UpdatePostContent},
        pages::PaginatedQuery,
    },
    State,
};

pub async fn create_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Json(req): Json<CreatePostDraft>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let mut user = match get_user_from_token!(jar, data, oauth::AppScope::UserCreateDrafts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    // award achievement
    if let Err(e) = data
        .add_achievement(&mut user, AchievementName::CreateDraft.into(), true)
        .await
    {
        return Json(e.into());
    }

    // ...
    match data
        .create_draft(PostDraft::new(req.content, user.id))
        .await
    {
        Ok(id) => Json(ApiReturn {
            ok: true,
            message: "Draft created".to_string(),
            payload: Some(id.to_string()),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn delete_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserDeleteDrafts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.delete_draft(id, user).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Draft deleted".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn update_content_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
    Json(req): Json<UpdatePostContent>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserEditDrafts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.update_draft_content(id, user, req.content).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Draft updated".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn get_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserReadDrafts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.get_draft_by_id(id).await {
        Ok(d) => {
            if d.owner != user.id {
                return Json(Error::NotAllowed.into());
            }

            Json(ApiReturn {
                ok: true,
                message: "Success".to_string(),
                payload: Some(d),
            })
        }
        Err(e) => Json(e.into()),
    }
}

pub async fn get_drafts_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Query(props): Query<PaginatedQuery>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserReadDrafts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.get_drafts_by_user(user.id, 12, props.page).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Success".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn get_all_drafts_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserReadDrafts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.get_drafts_by_user_all(user.id).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Success".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}
