use crate::{
    get_user_from_token,
    image::{save_webp_buffer, JsonMultipart},
    routes::{
        api::v1::{
            communities::posts::MAXIMUM_FILE_SIZE, CreateProduct, UpdateProductDescription,
            UpdateProductName, UpdateProductPrice,
        },
        pages::PaginatedQuery,
    },
    State,
};
use axum::{
    extract::{Path, Query},
    response::IntoResponse,
    Extension, Json,
};
use crate::cookie::CookieJar;
use tetratto_core::model::{
    oauth,
    products::Product,
    uploads::{MediaType, MediaUpload},
    ApiReturn, Error,
};

pub async fn get_request(
    Path(id): Path<usize>,
    Extension(data): Extension<State>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    match data.get_product_by_id(id).await {
        Ok(x) => Json(ApiReturn {
            ok: true,
            message: "Success".to_string(),
            payload: Some(x),
        }),
        Err(e) => return Json(e.into()),
    }
}

pub async fn list_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Query(props): Query<PaginatedQuery>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserReadProducts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.get_products_by_user(user.id, 12, props.page).await {
        Ok(x) => Json(ApiReturn {
            ok: true,
            message: "Success".to_string(),
            payload: Some(x),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn create_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    JsonMultipart(uploads, req): JsonMultipart<CreateProduct>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserCreateProducts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    if uploads.len() > 4 {
        return Json(
            Error::MiscError("Too many uploads. Please use a maximum of 4".to_string()).into(),
        );
    }

    let mut product = Product::new(
        user.id,
        req.name,
        req.description,
        req.price,
        req.product_type,
    );

    // check sizes
    for img in &uploads {
        if img.len() > MAXIMUM_FILE_SIZE {
            return Json(Error::FileTooLarge.into());
        }
    }

    // create uploads
    for _ in 0..uploads.len() {
        product.uploads.push(
            match data
                .create_upload(MediaUpload::new(MediaType::Webp, product.owner))
                .await
            {
                Ok(u) => u.id,
                Err(e) => return Json(e.into()),
            },
        );
    }

    let product_uploads = product.uploads.clone();
    match data.create_product(product).await {
        Ok(x) => {
            // store uploads
            for (i, upload_id) in product_uploads.iter().enumerate() {
                let image = match uploads.get(i) {
                    Some(img) => img,
                    None => {
                        if let Err(e) = data.delete_upload(*upload_id).await {
                            return Json(e.into());
                        }

                        continue;
                    }
                };

                let upload = match data.get_upload_by_id(*upload_id).await {
                    Ok(u) => u,
                    Err(e) => return Json(e.into()),
                };

                if let Err(e) =
                    save_webp_buffer(&upload.path(&data.0.0).to_string(), image.to_vec(), None)
                {
                    return Json(Error::MiscError(e.to_string()).into());
                }
            }

            // ...
            Json(ApiReturn {
                ok: true,
                message: "Product created".to_string(),
                payload: x.id.to_string(),
            })
        }
        Err(e) => Json(e.into()),
    }
}

pub async fn update_name_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
    Json(req): Json<UpdateProductName>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserManageProducts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.update_product_name(id, &user, &req.name).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Product updated".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn update_description_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
    Json(req): Json<UpdateProductDescription>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserManageProducts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data
        .update_product_description(id, &user, &req.description)
        .await
    {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Product updated".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn update_price_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
    Json(req): Json<UpdateProductPrice>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserManageProducts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.update_product_price(id, &user, req.price).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Product updated".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn delete_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserManageProducts) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.delete_product(id, &user).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Product deleted".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}
