use crate::{
    get_user_from_token,
    routes::api::v1::{
        CreateService, UpdateServiceFileContent, UpdateServiceFiles, UpdateServiceName,
    },
    State,
};
use axum::{extract::Path, response::IntoResponse, Extension, Json};
use crate::cookie::CookieJar;
use tetratto_core::model::{auth::AchievementName, littleweb::Service, oauth, ApiReturn, Error};
use tetratto_shared::unix_epoch_timestamp;

pub async fn get_request(
    Path(id): Path<usize>,
    Extension(data): Extension<State>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    match data.get_service_by_id(id).await {
        Ok(x) => Json(ApiReturn {
            ok: true,
            message: "Success".to_string(),
            payload: Some(x),
        }),
        Err(e) => return Json(e.into()),
    }
}

pub async fn list_request(jar: CookieJar, Extension(data): Extension<State>) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserReadServices) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.get_services_by_user(user.id).await {
        Ok(x) => Json(ApiReturn {
            ok: true,
            message: "Success".to_string(),
            payload: Some(x),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn create_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Json(req): Json<CreateService>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let mut user = match get_user_from_token!(jar, data, oauth::AppScope::UserCreateServices) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    // award achievement
    if let Err(e) = data
        .add_achievement(&mut user, AchievementName::CreateSite.into(), true)
        .await
    {
        return Json(e.into());
    }

    // ...
    match data.create_service(Service::new(req.name, user.id)).await {
        Ok(x) => Json(ApiReturn {
            ok: true,
            message: "Service created".to_string(),
            payload: x.id.to_string(),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn update_name_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
    Json(req): Json<UpdateServiceName>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserManageServices) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.update_service_name(id, &user, &req.name).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Service updated".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn update_files_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
    Json(req): Json<UpdateServiceFiles>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserManageServices) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    let mut service = match data.get_service_by_id(id).await {
        Ok(x) => x,
        Err(e) => return Json(e.into()),
    };

    if req.id_path.is_empty() {
        service.files = req.files;
    } else {
        match service.file_mut(req.id_path) {
            Some(f) => f.children = req.files,
            None => return Json(Error::GeneralNotFound("file".to_string()).into()),
        }
    }

    match data.update_service_files(id, &user, service.files).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Service updated".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}

pub async fn update_content_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
    Json(req): Json<UpdateServiceFileContent>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserManageServices) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    let mut service = match data.get_service_by_id(id).await {
        Ok(x) => x,
        Err(e) => return Json(e.into()),
    };

    // update
    let file = match service.file_mut(req.id_path) {
        Some(f) => f,
        None => return Json(Error::GeneralNotFound("file".to_string()).into()),
    };

    file.content = req.content;

    // ...
    match data.update_service_files(id, &user, service.files).await {
        Ok(_) => match data
            .update_service_revision(id, unix_epoch_timestamp() as i64)
            .await
        {
            Ok(_) => Json(ApiReturn {
                ok: true,
                message: "Service updated".to_string(),
                payload: (),
            }),
            Err(e) => Json(e.into()),
        },
        Err(e) => Json(e.into()),
    }
}

pub async fn delete_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
) -> impl IntoResponse {
    let data = &(data.read().await).0;
    let user = match get_user_from_token!(jar, data, oauth::AppScope::UserManageServices) {
        Some(ua) => ua,
        None => return Json(Error::NotAllowed.into()),
    };

    match data.delete_service(id, &user).await {
        Ok(_) => Json(ApiReturn {
            ok: true,
            message: "Service deleted".to_string(),
            payload: (),
        }),
        Err(e) => Json(e.into()),
    }
}
