use super::render_error;
use crate::{assets::initial_context, get_lang, get_user_from_token, State};
use axum::{
    response::{Html, IntoResponse},
    extract::Path,
    Extension,
};
use crate::cookie::CookieJar;
use tetratto_core::model::{apps::AppData, permissions::FinePermission, Error};

/// `/developer`
pub async fn home_request(jar: CookieJar, Extension(data): Extension<State>) -> impl IntoResponse {
    let data = data.read().await;
    let user = match get_user_from_token!(jar, data.0) {
        Some(ua) => ua,
        None => {
            return Err(Html(
                render_error(Error::NotAllowed, &jar, &data, &None).await,
            ));
        }
    };

    let list = match data.0.get_apps_by_owner(user.id).await {
        Ok(p) => p,
        Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
    };

    let lang = get_lang!(jar, data.0);
    let mut context = initial_context(&data.0.0.0, lang, &Some(user)).await;
    context.insert("list", &list);

    // return
    Ok(Html(
        data.1.render("developer/home.html", &context).unwrap(),
    ))
}

/// `/developer/app/{id}`
pub async fn app_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
) -> impl IntoResponse {
    let data = data.read().await;
    let user = match get_user_from_token!(jar, data.0) {
        Some(ua) => ua,
        None => {
            return Err(Html(
                render_error(Error::NotAllowed, &jar, &data, &None).await,
            ));
        }
    };

    let app = match data.0.get_app_by_id(id).await {
        Ok(p) => p,
        Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
    };

    if user.id != app.owner && !user.permissions.check(FinePermission::MANAGE_APPS) {
        return Err(Html(
            render_error(Error::NotAllowed, &jar, &data, &None).await,
        ));
    }

    let data_limit = AppData::user_limit(&user);

    let lang = get_lang!(jar, data.0);
    let mut context = initial_context(&data.0.0.0, lang, &Some(user)).await;

    context.insert("app", &app);
    context.insert("data_limit", &data_limit);

    // return
    Ok(Html(data.1.render("developer/app.html", &context).unwrap()))
}

/// `/auth/connections_link/app/{id}`
pub async fn connection_callback_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
) -> impl IntoResponse {
    let data = data.read().await;
    let user = match get_user_from_token!(jar, data.0) {
        Some(ua) => ua,
        None => return Html(render_error(Error::NotAllowed, &jar, &data, &None).await),
    };

    let app = match data.0.get_app_by_id(id).await {
        Ok(p) => p,
        Err(e) => return Html(render_error(e, &jar, &data, &Some(user)).await),
    };

    let lang = get_lang!(jar, data.0);
    let mut context = initial_context(&data.0.0.0, lang, &Some(user)).await;
    context.insert("app", &app);

    Html(data.1.render("developer/link.html", &context).unwrap())
}
