pub mod auth;
pub mod chats;
pub mod communities;
pub mod developer;
pub mod forge;
pub mod journals;
pub mod littleweb;
pub mod marketplace;
pub mod misc;
pub mod mod_panel;
pub mod profile;
pub mod stacks;

use axum::{
    routing::{get, post},
    Router,
};
use crate::cookie::CookieJar;
use serde::Deserialize;
use tetratto_core::{
    model::{Error, auth::User},
};

use crate::{assets::initial_context, get_lang, InnerState};

pub fn routes() -> Router {
    Router::new()
        // timelines
        .route("/", get(misc::index_request))
        .route("/popular", get(misc::popular_request))
        .route("/following", get(misc::following_request))
        .route("/all", get(misc::all_request))
        .route("/search", get(misc::search_request))
        .route(
            "/_swiss_army_timeline",
            get(misc::swiss_army_timeline_request),
        )
        // question timelines
        .route("/questions", get(misc::index_questions_request))
        .route("/popular/questions", get(misc::popular_questions_request))
        .route(
            "/following/questions",
            get(misc::following_questions_request),
        )
        .route("/all/questions", get(misc::all_questions_request))
        // misc
        .route("/notifs", get(misc::notifications_request))
        .route("/requests", get(misc::requests_request))
        .route("/achievements", get(misc::achievements_request))
        .route("/doc/{*file_name}", get(misc::markdown_document_request))
        .fallback_service(get(misc::not_found))
        // mod
        .route("/mod_panel/audit_log", get(mod_panel::audit_log_request))
        .route("/mod_panel/reports", get(mod_panel::reports_request))
        .route(
            "/mod_panel/file_report",
            get(mod_panel::file_report_request),
        )
        .route("/mod_panel/ip_bans", get(mod_panel::ip_bans_request))
        .route(
            "/mod_panel/profile/{id}",
            get(mod_panel::manage_profile_request),
        )
        .route(
            "/mod_panel/profile/{id}/warnings",
            get(mod_panel::manage_profile_warnings_request),
        )
        .route("/mod_panel/stats", get(mod_panel::stats_request))
        // auth
        .route("/auth/register", get(auth::register_request))
        .route("/auth/login", get(auth::login_request))
        .route(
            "/auth/connections_link/{service}",
            get(auth::connection_callback_request),
        )
        .route(
            "/auth/connections_link/app/{id}",
            get(developer::connection_callback_request),
        )
        .route(
            "/auth/connections_link/seller/reload",
            get(marketplace::connection_reload_request),
        )
        .route(
            "/auth/connections_link/seller/return",
            get(marketplace::connection_return_request),
        )
        // profile
        .route("/settings", get(profile::settings_request))
        .route("/@{username}", get(profile::posts_request))
        .route("/@{username}/media", get(profile::media_request))
        .route("/@{username}/outbox", get(profile::outbox_request))
        .route("/@{username}/replies", get(profile::replies_request))
        .route("/@{username}/following", get(profile::following_request))
        .route("/@{username}/followers", get(profile::followers_request))
        // communities
        .route("/communities", get(communities::list_request))
        .route("/communities/search", get(communities::search_request))
        .route(
            "/communities/intents/post",
            get(communities::create_post_request),
        )
        .route("/community/{title}", get(communities::feed_request))
        .route(
            "/community/{title}/questions",
            get(communities::questions_request),
        )
        .route("/community/{id}/manage", get(communities::settings_request))
        .route(
            "/community/{title}/members",
            get(communities::members_request),
        )
        .route("/post/{id}", get(communities::post_request))
        .route("/post/{id}/reposts", get(communities::reposts_request))
        .route("/post/{id}/likes", get(communities::likes_request))
        .route("/question/{id}", get(communities::question_request))
        // chats
        .route("/chats", get(chats::redirect_request))
        .route("/chats/{community}/{channel}", get(chats::app_request))
        .route(
            "/chats/{community}/{channel}/_stream",
            get(chats::stream_request),
        )
        .route(
            "/chats/{community}/{channel}/_render",
            post(chats::message_request),
        )
        .route(
            "/chats/{community}/{channel}/_channels",
            get(chats::channels_request),
        )
        // forge
        .route("/forges", get(forge::home_request))
        .route("/forge/{title}", get(forge::info_request))
        .route("/forge/{title}/tickets", get(forge::tickets_request))
        .route("/forge/{title}/members", get(communities::members_request))
        // developer
        .route("/developer", get(developer::home_request))
        .route("/developer/app/{id}", get(developer::app_request))
        // stacks
        .route("/stacks", get(stacks::list_request))
        .route("/stacks/{id}", get(stacks::feed_request))
        .route("/stacks/{id}/manage", get(stacks::manage_request))
        .route("/stacks/add_user/{id}", get(stacks::add_user_request))
        // journals
        .route("/journals", get(journals::redirect_request))
        .route("/journals/{journal}/{note}", get(journals::app_request))
        .route("/@{owner}/{journal}", get(journals::index_view_request))
        .route("/@{owner}/{journal}/{note}", get(journals::view_request))
        .route("/x/{note}", get(journals::global_view_request))
        // littleweb
        .route("/services", get(littleweb::services_request))
        .route("/domains", get(littleweb::domains_request))
        .route("/services/{id}", get(littleweb::service_request))
        .route("/domains/{id}", get(littleweb::domain_request))
        .route("/net", get(littleweb::browser_home_request))
        .route("/net/{*uri}", get(littleweb::browser_request))
        // marketplace
        .route(
            "/settings/seller",
            get(marketplace::seller_settings_request),
        )
}

pub fn lw_routes() -> Router {
    Router::new()
}

pub async fn render_error(
    e: Error,
    jar: &CookieJar,
    data: &InnerState,
    user: &Option<User>,
) -> String {
    let lang = get_lang!(jar, data.0);
    let mut context = initial_context(&data.0.0.0, lang, user).await;
    context.insert("error_text", &e.to_string());
    data.1.render("misc/error.html", &context).unwrap()
}

#[derive(Deserialize)]
pub struct PaginatedQuery {
    #[serde(default)]
    pub page: usize,
    #[serde(default)]
    pub before: usize,
}

#[derive(Deserialize)]
pub struct ChatsAppQuery {
    #[serde(default)]
    pub page: usize,
    #[serde(default)]
    pub nav: bool,
    #[serde(default)]
    pub message: usize,
}

#[derive(Deserialize)]
pub struct ProfileQuery {
    #[serde(default)]
    pub page: usize,
    #[serde(default)]
    pub warning: bool,
    #[serde(default)]
    pub tag: String,
    #[serde(default, alias = "r")]
    pub responses_only: bool,
    #[serde(default, alias = "f")]
    pub force: bool,
}

#[derive(Deserialize)]
pub struct SearchedQuery {
    #[serde(default)]
    pub text: String,
    #[serde(default)]
    pub page: usize,
}

#[derive(Deserialize)]
pub struct RepostsQuery {
    #[serde(default)]
    pub quotes: bool,
    #[serde(default)]
    pub page: usize,
}

#[derive(Deserialize)]
pub struct JournalsAppQuery {
    #[serde(default)]
    pub view: bool,
    #[serde(default)]
    pub tag: String,
}
