use axum::{
    extract::{Path, Query},
    response::{Html, IntoResponse},
    Extension,
};
use crate::cookie::CookieJar;
use tetratto_core::model::{
    auth::User,
    permissions::FinePermission,
    stacks::{StackMode, StackPrivacy},
    Error,
};
use crate::{assets::initial_context, get_lang, get_user_from_token, State};
use super::{render_error, PaginatedQuery};

/// `/stacks`
pub async fn list_request(jar: CookieJar, Extension(data): Extension<State>) -> impl IntoResponse {
    let data = data.read().await;
    let user = match get_user_from_token!(jar, data.0) {
        Some(ua) => ua,
        None => {
            return Err(Html(
                render_error(Error::NotAllowed, &jar, &data, &None).await,
            ));
        }
    };

    let list = match data.0.get_stacks_by_user(user.id).await {
        Ok(p) => p,
        Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
    };

    let lang = get_lang!(jar, data.0);
    let mut context = initial_context(&data.0.0.0, lang, &Some(user)).await;

    context.insert("list", &list);

    // return
    Ok(Html(data.1.render("stacks/list.html", &context).unwrap()))
}

/// `/stacks/{id}`
pub async fn feed_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
    Query(req): Query<PaginatedQuery>,
) -> impl IntoResponse {
    let data = data.read().await;
    let user = match get_user_from_token!(jar, data.0) {
        Some(ua) => ua,
        None => {
            return Err(Html(
                render_error(Error::NotAllowed, &jar, &data, &None).await,
            ));
        }
    };

    let stack = match data.0.get_stack_by_id(id).await {
        Ok(s) => s,
        Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
    };

    if stack.privacy == StackPrivacy::Private
        && user.id != stack.owner
        && !(stack.mode == StackMode::Circle && stack.users.contains(&user.id))
        && !user.permissions.check(FinePermission::MANAGE_STACKS)
    {
        return Err(Html(
            render_error(Error::NotAllowed, &jar, &data, &None).await,
        ));
    }

    let lang = get_lang!(jar, data.0);
    let mut context = initial_context(&data.0.0.0, lang, &Some(user.clone())).await;

    context.insert("page", &req.page);
    context.insert("stack", &stack);

    if stack.mode == StackMode::BlockList {
        let list = match data.0.get_stack_users(stack.id, 12, req.page).await {
            Ok(l) => l,
            Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
        };

        context.insert("list", &list);
        context.insert(
            "is_blocked",
            &data
                .0
                .get_stackblock_by_initiator_stack(user.id, stack.id)
                .await
                .is_ok(),
        );
    }

    // return
    Ok(Html(data.1.render("stacks/feed.html", &context).unwrap()))
}

/// `/stacks/{id}/manage`
pub async fn manage_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
) -> impl IntoResponse {
    let data = data.read().await;
    let user = match get_user_from_token!(jar, data.0) {
        Some(ua) => ua,
        None => {
            return Err(Html(
                render_error(Error::NotAllowed, &jar, &data, &None).await,
            ));
        }
    };

    let stack = match data.0.get_stack_by_id(id).await {
        Ok(s) => s,
        Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
    };

    if user.id != stack.owner && !user.permissions.check(FinePermission::MANAGE_STACKS) {
        return Err(Html(
            render_error(Error::NotAllowed, &jar, &data, &None).await,
        ));
    }

    let mut new_users = stack.users.clone();
    let mut changed_stack_users: bool = false;

    let mut users: Vec<User> = Vec::new();

    for uid in &stack.users {
        users.push(match data.0.get_user_by_id(uid.to_owned()).await {
            Ok(ua) => ua,
            Err(_) => {
                // user deleted profile, remove from list
                new_users.remove(stack.users.iter().position(|x| x == uid).unwrap());
                changed_stack_users = true;
                continue;
            }
        });
    }

    if changed_stack_users {
        if let Err(e) = data.0.update_stack_users(stack.id, &user, new_users).await {
            return Err(Html(render_error(e, &jar, &data, &None).await));
        }
    }

    let lang = get_lang!(jar, data.0);
    let mut context = initial_context(&data.0.0.0, lang, &Some(user)).await;

    context.insert("stack", &stack);
    context.insert("users", &users);

    // return
    Ok(Html(data.1.render("stacks/manage.html", &context).unwrap()))
}

/// `/stacks/add_user`
pub async fn add_user_request(
    jar: CookieJar,
    Path(id): Path<usize>,
    Extension(data): Extension<State>,
) -> impl IntoResponse {
    let data = data.read().await;
    let user = match get_user_from_token!(jar, data.0) {
        Some(ua) => ua,
        None => {
            return Err(Html(
                render_error(Error::NotAllowed, &jar, &data, &None).await,
            ));
        }
    };

    let add_user = match data.0.get_user_by_id(id).await {
        Ok(ua) => ua,
        Err(e) => return Err(Html(render_error(e, &jar, &data, &None).await)),
    };

    let stacks = match data.0.get_stacks_by_user(user.id).await {
        Ok(p) => p,
        Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
    };

    let lang = get_lang!(jar, data.0);
    let mut context = initial_context(&data.0.0.0, lang, &Some(user)).await;

    context.insert("stacks", &stacks);
    context.insert("add_user", &add_user);

    // return
    Ok(Html(
        data.1.render("stacks/add_user.html", &context).unwrap(),
    ))
}
